package metaobj;

import java.util.List;
import meta.AnnotationArgumentsKind;
import meta.AttachedDeclarationKind;
import meta.CyanMetaobject;
import meta.CyanMetaobjectWithAt;
import meta.IAction_afti;
import meta.ICheckDeclaration_afsa;
import meta.ICompiler_afti;
import meta.ICompiler_dsa;
import meta.IDeclaration;
import meta.ISlotInterface;
import meta.Tuple2;
import meta.WrASTVisitor;
import meta.WrCyanMetaobjectAnnotation;
import meta.WrCyanMetaobjectWithAtAnnotation;
import meta.WrEnv;
import meta.WrProgramUnit;
import meta.WrStatementBreak;

public class CyanMetaobjectTwo extends CyanMetaobjectWithAt
	implements IAction_afti, ICheckDeclaration_afsa {

    public CyanMetaobjectTwo() {
        super("two", AnnotationArgumentsKind.ZeroOrMoreParameters,
                new AttachedDeclarationKind[] {
                AttachedDeclarationKind.PROTOTYPE_DEC
        });
    }

	@Override
	public Tuple2<StringBuffer, String> afti_codeToAdd(
			ICompiler_afti compiler,
			List<Tuple2<WrCyanMetaobjectAnnotation,
			List<ISlotInterface>>> infoList) {

		WrCyanMetaobjectWithAtAnnotation annot =
				this.getMetaobjectAnnotation();
		List<Object> paramList = annot.getJavaParameterList();
		if ( paramList.size() < 2 || paramList.size() > 3 ) {
			this.addError("This annotation should have two or "
					+ "three parameters");
		}

		if ( !(paramList.get(0) instanceof String) ||
				!(paramList.get(1) instanceof String) ) {
			this.addError("The two first parameters to this annotation "
					+ "should be strings");
		}
		return new Tuple2<>(
				new StringBuffer( CyanMetaobject.removeQuotes( (String ) paramList.get(0)) ),
				CyanMetaobject.removeQuotes( (String ) paramList.get(1)) );
	}


    @Override
    public void afsa_checkDeclaration(ICompiler_dsa compiler) {

        IDeclaration dec = this.getAttachedDeclaration();
        WrProgramUnit pu = (WrProgramUnit ) dec;
        WrEnv env = compiler.getEnv();


		WrCyanMetaobjectWithAtAnnotation annot =
				this.getMetaobjectAnnotation();
		List<Object> paramList = annot.getJavaParameterList();

		boolean checkBreak = false;

		if ( paramList.size() == 3 ) {
			if ( !(paramList.get(2) instanceof String) ) {
				this.addError("The third parameter to this annotation "
						+ "should be a string");
			}
			String strBreak = (String ) paramList.get(2);
			/*
			 * the parameter may be give as an identifer, without
			 * quotes, or with quotes. Both are considered strings.
			       @two( "var Int count = 0;",
      				     "var Int count",
      				     break )   // no quotes

			       @two( "var Int count = 0;",
      				     "var Int count",
      				     "break" )   // with quotes

			   So, if there are quotes, strBreak, in this case,
			   would be
			        "\"break\""
			   So let us remove the quotes, if they exist
			 */
			strBreak = CyanMetaobject.removeQuotes(strBreak);
			if ( strBreak.equalsIgnoreCase("break") ) {
				checkBreak = true;
			}
		}

		if ( checkBreak ) {
			/*
			 * we show TWO ways of checking if there is a 'break' statement
			 */

	        cyanruntime.Ref<Integer> breakCount = new cyanruntime.Ref<>();
		    breakCount.elem = 0;

			pu.accept( new WrASTVisitor() {
		    	@Override
				public void visit(WrStatementBreak node, WrEnv env2) {
		    		++breakCount.elem;
				}

			}, env);
			if ( breakCount.elem != 0 ) this.addError("'break' statement is not allowed");

			BreakCountASTVisitor bcount = new BreakCountASTVisitor();
			pu.accept(bcount, env);
			if ( bcount.breakCount != 0 )
				addError("'break' statement is not allowed");

			/*
			 * Of course, if checkBreak is true and there is a break,
			 * TWO error messages will be issued
			 */
		}

    }
}

class BreakCountASTVisitor extends WrASTVisitor {
    public int breakCount = 0;

    @Override
    public void visit(WrStatementBreak node, WrEnv env) {
        ++breakCount;
    }
}

